package sets;


public class OptimisticSet<T> extends SequentialSet<T>
{
    public OptimisticSet() {
        super();
    }

    // is `pred' reachable from `head' and pointing to `curr'?
    protected boolean valid(Node<T> pred, Node<T> curr) {
        Node<T> node = head;  // start from `head'
        while (node.key() <= pred.key()) { // does `pred' point to `curr'?
            if (node == pred) return pred.next() == curr;
            node = node.next(); // continue to the next node
        } // until node.pred > pred.key
        return false;
    }

    // find without locking, and then lock and validate
    // if validation successful, add while locking; otherwise unlock and repeat
    @Override
    public boolean add(T item) {
        // new node
        Node<T> node = newNode(item);
        do {
            Position<T> where = find(head, node.key());  // find without locking
            Node<T> pred = where.pred, curr = where.curr;
            pred.lock();
            curr.lock();  // lock position
            try {
                if(valid(pred, curr))  // if position still valid
                    return rawAdd(pred, curr, node); // add node as in SequentialSet, while locking
            } finally {
                pred.unlock();
                curr.unlock();
            }
        } while (true);  // if not valid, try again!
    }

    // find without locking, and then lock and validate
    // if validation successful, remove while locking; otherwise unlock and repeat
    @Override
    public boolean remove(T item) {
        int key = item.hashCode();
        do {
            Position<T> where = find(head, key);  // find without locking
            Node<T> pred = where.pred, curr = where.curr;
            pred.lock();
            curr.lock();  // lock position
            try {
                if(valid(pred, curr))  // if position still valid
                    return rawRemove(pred, curr, key); // remove node as in SequentialSet, while locking
            } finally {
                pred.unlock();
                curr.unlock();
            }
        } while (true);  // if not valid, try again!
    }

    // find without locking, and then lock and validate
    // if validation successful, has while locking; otherwise unlock and repeat
    @Override
    public boolean has(T item) {
        int key = item.hashCode();
        do {
            Position<T> where = find(head, key);  // find without locking
            Node<T> pred = where.pred, curr = where.curr;
            pred.lock();
            curr.lock();  // lock position
            try {
                if(valid(pred, curr))  // if position still valid
                    return rawHas(curr, key); // check node as in SequentialSet, while locking
            } finally {
                pred.unlock();
                curr.unlock();
            }
        } while (true);  // if not valid, try again!
    }

    // node whose chain can be navigated concurrently without locking
    protected Node<T> newNode(T item) {
        return new ReadWriteNode<>(item);
    }

    // node whose chain can be navigated concurrently without locking
    protected Node<T> newNode(int key) {
        return new ReadWriteNode<>(key);
    }
}
